#!/usr/bin/env python
#
#  annotations.py
"""
Add serialisation and deserialisation capability to attrs_ classes.

.. _attrs: https://www.attrs.org/en/stable/

Based on `attrs-serde <https://github.com/jondot/attrs-serde>`_.

Example usage
--------------

.. code-block:: python

	>>> import attr
	>>> from attr_utils.serialise import serde

	>>> person_dict = {"contact": {"personal": {"name": "John"}, "phone": "555-112233"}}

	>>> name_path = ["contact", "personal", "name"]
	>>> phone_path = ["contact", "phone"]

	>>> @serde
	... @attr.s
	... class Person(object):
	... 	name = attr.ib(metadata={"to": name_path, "from": name_path})
	... 	phone = attr.ib(metadata={"to": phone_path, "from": phone_path})

	>>> p = Person.from_dict(person_dict)
	Person(name=John phone=555-112233)

	>>> p.to_dict
	{"contact": {"personal": {"name": "John"}, "phone": "555-112233"}}


API Reference
---------------
"""

#  From https://github.com/jondot/attrs-serde
#  MIT License
#
#  Copyright (c) 2019 Dotan Nahum
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#

# stdlib
from typing import TYPE_CHECKING, Any, Callable, Mapping, MutableMapping, Optional, Type, Union, overload

try:
	# 3rd party
	from cytoolz.curried import filter, get_in, map, pipe, reduce, update_in  # type: ignore
except ImportError:
	from toolz.curried import get_in, update_in, reduce, pipe, map, filter  # type: ignore

# 3rd party
from attr import asdict, fields

__all__ = ["serde"]

if TYPE_CHECKING:
	AttrsClass = Any
else:
	# this package
	from attr_utils.utils import AttrsClass


@overload
def serde(
		cls: Type,
		from_key: str = ...,
		to_key: str = ...,
		) -> Type[AttrsClass]:
	...  # pragma: no cover


@overload
def serde(
		cls: None = None,
		from_key: str = ...,
		to_key: str = ...,
		) -> Callable[[Type[AttrsClass]], Type[AttrsClass]]:
	...  # pragma: no cover


def serde(
		cls: Optional[Type[AttrsClass]] = None,
		from_key: str = "from",
		to_key: str = "to",
		) -> Union[Type[AttrsClass], Callable[[Type[AttrsClass]], Type[AttrsClass]]]:
	r"""
	Decorator to add serialisation and deserialisation capabilities to attrs classes.

	The keys used in the dictionary output, and used when creating the class from a dictionary,
	can be controlled using the ``metadata`` argument in :func:`attr.ib`:

	.. code-block::

		from attr_utils.serialize import serde
		import attr

		@serde
		@attr.s
		class Person(object):
			name = attr.ib(metadata={"to": name_path, "from": name_path})
			phone = attr.ib(metadata={"to": phone_path, "from": phone_path})

	The names of the keys given in the ``metadata`` argument can be controlled with the
	``from_key`` and ``to_key`` arguments:

	.. code-block::

		from attr_utils.serialize import serde
		import attr

		@serde(from_key="get", to_key="set")
		@attr.s
		class Person(object):
			name = attr.ib(metadata={"get": name_path, "set": name_path})
			phone = attr.ib(metadata={"get": phone_path, "set": phone_path})


	This may be required when using other extensions to attrs.


	:param cls: The attrs class to add the methods to.
	:param from_key:
	:param to_key:

	:rtype:

	Classes decorated with :deco:`~attr_utils.serialise.serde` will have two new methods added:

	.. py:classmethod:: from_dict(d)

		Construct an instance of the class from a dictionary.

		:param d: The dictionary
		:type d: :class:`~typing.Mapping`\[:class:`str`, :py:obj:`~typing.Any`\]

	.. py:method:: to_dict(convert_values=False):

		Returns a dictionary containing the contents of the class.

		:param convert_values: Recurse into other attrs classes, and convert tuples, sets etc.
			into lists. This may be required to later construct a new class from the
			dictionary if the class uses complex converter functions.
		:type convert_values: :class:`bool`

		:rtype: :class:`~typing.MutableMapping`\[:class:`str`, :py:obj:`~typing.Any`\]

		.. versionchanged:: 0.5.0

			By default values are left unchanged. In version 0.4.0 these were converted
			to basic Python types, which may be undesirable. The original behaviour can be
			restored using the ``convert_values`` parameter.

	"""

	def serde_with_class(cls: Type[AttrsClass]) -> Type[AttrsClass]:

		def from_dict(cls, d: Mapping[str, Any]):
			from_fields = list(map(lambda a: (a, get_in([from_key], a.metadata, [a.name])), fields(cls)))

			return cls(**dict(map(
					lambda f: (f[0].name, get_in(f[1], d, f[0].default)),
					from_fields,
					)))

		def to_dict(self, convert_values: bool = False) -> MutableMapping[str, Any]:
			to_fields = pipe(
					fields(self.__class__),
					map(lambda a: (a, get_in([to_key], a.metadata))),
					filter(lambda f: f[1]),
					list,
					)

			if convert_values:
				d = asdict(self)
			else:
				d = {a.name: getattr(self, a.name) for a in fields(self.__class__)}

			if not to_fields:
				return d

			return reduce(
					lambda acc, f: update_in(acc, f[1], lambda _: d[f[0].name]),
					to_fields,
					{},
					)

		from_dict.__doc__ = f"""
		Construct an instance of :class:`~.{cls.__name__}` from a dictionary.

		:param d: The dictionary
		"""
		from_dict.__qualname__ = f"{cls.__name__}.from_dict"
		from_dict.__module__ = cls.__module__
		cls.from_dict = classmethod(from_dict)

		to_dict.__doc__ = f"""
Returns a dictionary containing the contents of the :class:`~.{cls.__name__}` object.

:param convert_values: Recursively convert values into dictionaries, lists etc. as appropriate.
"""
		to_dict.__qualname__ = f"{cls.__name__}.to_dict"
		to_dict.__module__ = cls.__module__
		cls.to_dict = to_dict

		return cls

	if cls is not None:
		return serde_with_class(cls)
	else:
		return serde_with_class



#!/usr/bin/env python
#
#  docstrings.py
"""
Add better docstrings to attrs generated functions.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import re
from typing import Optional, Pattern, Type

# 3rd party
from domdf_python_tools.doctools import PYPY, base_new_docstrings, prettify_docstrings
from domdf_python_tools.typing import MethodDescriptorType, MethodWrapperType, WrapperDescriptorType

__all__ = ["add_attrs_doc"]

attrs_docstring = "Automatically created by attrs."
ne_default = "Check equality and either forward a NotImplemented or return the result\n    negated."
attrs_20_1_docstring: Pattern = re.compile(r"^\s*Method generated by attrs for class .*\.\s*")
repr_doc_template = "Return a string representation of the :class:`~.{name}`."
pickle_state_template = "Used for `pickling <https://docs.python.org/3/library/pickle.html>`_.\n\n{doc}"


def add_attrs_doc(obj: Type) -> Type:
	"""
	Add better docstrings to attrs generated functions.

	:param obj: The class to improve the docstrings for.
	"""

	obj = prettify_docstrings(obj)

	new_docstrings = {
			**base_new_docstrings,
			"__repr__": repr_doc_template.format(name=obj.__name__),
			"__setstate__": pickle_state_template.format(doc=attrs_docstring),
			"__getstate__": pickle_state_template.format(doc=attrs_docstring),
			}

	if hasattr(obj, "__ne__"):
		if PYPY or not isinstance(obj.__ne__, (WrapperDescriptorType, MethodDescriptorType, MethodWrapperType)):
			if obj.__ne__.__doc__ is None or obj.__ne__.__doc__.strip() in {object.__ne__.__doc__, ne_default}:
				obj.__ne__.__doc__ = new_docstrings["__ne__"]

	if hasattr(obj, "__repr__"):
		if (
				obj.__repr__.__doc__ is None or obj.__repr__.__doc__.strip() == attrs_docstring
				or attrs_20_1_docstring.match(obj.__repr__.__doc__)
				):
			_new_doc = f"{new_docstrings['__repr__']}"  # \n\n{attrs_docstring}
			obj.__repr__.__doc__ = _new_doc  # prevents strange formatting in pycharm

	for attr_name in new_docstrings:

		if not hasattr(obj, attr_name):
			continue

		attribute = getattr(obj, attr_name)

		if not PYPY and isinstance(attribute, (WrapperDescriptorType, MethodDescriptorType, MethodWrapperType)):
			continue

		if attribute is None:
			continue

		doc: Optional[str] = attribute.__doc__

		if doc is None or doc.strip() == attrs_docstring or attrs_20_1_docstring.match(doc):
			_new_doc = f"{new_docstrings[attr_name]}"  # \n\n{attrs_docstring}
			attribute.__doc__ = _new_doc  # prevents strange formatting in pycharm

	return obj



#!/usr/bin/env python3
#
#  autoattrs.py
"""
A Sphinx directive for documenting `attrs <https://www.attrs.org/>`_ classes.

.. attention::

	Due to changes in the :mod:`typing` module :mod:`~attr_utils.autoattrs` is only officially supported on
	Python 3.7 and above.

.. rst:directive:: autoattrs

	Autodoc directive to document an `attrs <https://www.attrs.org/>`_ class.

	It behaves much like :rst:dir:`autoclass`. It can be used directly or as part of :rst:dir:`automodule`.

	Docstrings for parameters in ``__init__`` can be given in the class docstring or alongside each attribute
	(see :rst:dir:`autoattribute` for the syntax). The second option is recommended as it interacts better
	with other parts of autodoc. However, the class docstring can be used to override the description
	for a given parameter.

	.. rst:directive:option:: no-init-attribs
		:type: flag

		Excludes attributes taken as arguments in ``__init__`` from the output, even if they are documented.

		This may be useful for simple classes where converter functions aren't used.

		This option cannot be used as part of :rst:dir:`automodule`.


.. versionadded:: 0.1.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Parts based on https://github.com/sphinx-doc/sphinx
#  |  Copyright (c) 2007-2020 by the Sphinx team (see AUTHORS file).
#  |  BSD Licensed
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are
#  |  met:
#  |
#  |  * Redistributions of source code must retain the above copyright
#  |   notice, this list of conditions and the following disclaimer.
#  |
#  |  * Redistributions in binary form must reproduce the above copyright
#  |   notice, this list of conditions and the following disclaimer in the
#  |   documentation and/or other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  |  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  |  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  |  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  |  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  |  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  |  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  |  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  |  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  |  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from textwrap import dedent
from typing import Any, Dict, List, MutableMapping, Optional, Tuple, Type

# 3rd party
import attr
from sphinx.application import Sphinx  # nodep
from sphinx.ext.autodoc import ClassDocumenter, Documenter  # nodep
from sphinx.pycode import ModuleAnalyzer  # nodep
from sphinx_toolbox import __version__  # nodep
from sphinx_toolbox.more_autosummary import PatchedAutoSummClassDocumenter  # nodep
from sphinx_toolbox.utils import Param, flag, parse_parameters, unknown_module_warning  # nodep

# this package
from attr_utils.docstrings import add_attrs_doc
from attr_utils.utils import AttrsClass

__all__ = ["AttrsDocumenter", "setup"]


class AttrsDocumenter(PatchedAutoSummClassDocumenter):
	r"""
	Sphinx autodoc :class:`~sphinx.ext.autodoc.Documenter`
	for documenting `attrs <https://www.attrs.org/>`__ classes.

	.. versionadded:: 0.1.0

	.. versionchanged:: 0.3.0

		Parameters for ``__init__`` can be documented either in the class docstring or for the attribute.
		The class docstring has priority.

	.. versionchanged:: 0.3.0

		Added support for `autodocsumm <https://github.com/Chilipp/autodocsumm>`_.
	"""  # noqa: D400

	objtype = "attrs"
	directivetype = "class"
	priority = ClassDocumenter.priority + 1
	option_spec = {
			**PatchedAutoSummClassDocumenter.option_spec,
			"no-init-attribs": flag,
			}
	object: Type[AttrsClass]  # noqa: A003  # pylint: disable=redefined-builtin

	@classmethod
	def can_document_member(cls, member: Any, membername: str, isattr: bool, parent: Any) -> bool:
		"""
		Called to see if a member can be documented by this documenter.

		:param member:
		:param membername:
		:param isattr:
		:param parent:
		"""

		return attr.has(member) and isinstance(member, type)

	def add_content(self, more_content: Any, no_docstring: bool = False) -> None:  # type: ignore
		"""
		Add extra content (from docstrings, attribute docs etc.), but not the class docstring.

		:param more_content:
		:param no_docstring:
		"""

		Documenter.add_content(self, more_content, True)

		# set sourcename and add content from attribute documentation
		sourcename = self.get_sourcename()

		params, pre_output, post_output = self._get_docstring()

		self.add_line('', sourcename)
		for line in list(self.process_doc([pre_output])):
			self.add_line(line, sourcename)
		self.add_line('', sourcename)

	def _get_docstring(self) -> Tuple[Dict[str, Param], List[str], List[str]]:
		"""
		Returns params, pre_output, post_output.
		"""

		# Size varies depending on docutils config
		tab_size = self.env.app.config.docutils_tab_width

		if self.object.__doc__:
			docstring = dedent(self.object.__doc__).expandtabs(tab_size).split('\n')
		else:
			docstring = []

		return parse_parameters(docstring, tab_size=tab_size)

	def import_object(self, raiseerror: bool = False) -> bool:
		"""
		Import the object given by ``self.modname`` and ``self.objpath`` and set it as ``self.object``.

		If the object is an `attrs <https://www.attrs.org/>`__ class
		:func:`attr_utils.docstrings.add_attrs_doc` will be called.

		:param raiseerror:

		:return: :py:obj:`True` if successful, :py:obj:`False` if an error occurred.
		"""

		ret = super().import_object(raiseerror)
		if attr.has(self.object):
			self.object = add_attrs_doc(self.object)

		return ret

	def sort_members(
			self,
			documenters: List[Tuple[Documenter, bool]],
			order: str,
			) -> List[Tuple[Documenter, bool]]:
		"""
		Sort the given member list and add attribute docstrings to the class docstring.

		:param documenters:
		:param order:
		"""

		# The documenters for the fields and methods, in the desired order
		# The fields will be in bysource order regardless of the order option
		documenters = super().sort_members(documenters, order)

		if hasattr(self, "_docstring_processed"):
			return documenters

		# Mapping of member names to docstrings (as list of strings)
		member_docstrings = {
				k[1]: v
				for k, v in ModuleAnalyzer.for_module(self.object.__module__).find_attr_docs().items()
				}

		# set sourcename and add content from attribute documentation
		sourcename = self.get_sourcename()

		parameter_docs = []
		params, pre_output, post_output = self._get_docstring()
		all_docs = {}

		for field in (a.name for a in attr.fields(self.object) if a.init):
			doc: List[str] = ['']

			# Prefer doc from class docstring
			if field in params:
				doc, arg_type = params.pop(field).values()  # type: ignore

			# Otherwise use attribute docstring
			if not ''.join(doc).strip() and field in member_docstrings:
				doc = member_docstrings[field]

			field_entry = [f":param {field}:", *doc]
			parameter_docs.append(' '.join(field_entry))
			all_docs[field] = ''.join(doc).strip()

		self.add_line('', sourcename)
		for line in list(self.process_doc([[*parameter_docs, '', '', *post_output]])):
			self.add_line(line, sourcename)
		self.add_line('', sourcename)

		self._docstring_processed = True

		if hasattr(self.object, "__slots__"):
			slots_dict: MutableMapping[str, Optional[str]] = {}
			for item in self.object.__slots__:
				if item in all_docs:
					slots_dict[item] = all_docs[item]
				else:
					slots_dict[item] = None

			self.object.__slots__ = slots_dict

		if hasattr(self, "add_autosummary"):
			self.add_autosummary()

		# import functools
		#
		# for documenter in documenters:
		# 	documenter[0].parse_name()
		# 	if documenter[0].objpath[-1] in all_docs:
		# 		def get_doc(encoding: str = None, ignore: int = None, doc: str = '') -> List[List[str]]:
		# 			return [[doc]]
		# 		documenter[0].get_doc = functools.partial(get_doc, doc=all_docs[documenter[0].objpath[-1]])

		return documenters

	def filter_members(
			self,
			members: List[Tuple[str, Any]],
			want_all: bool,
			) -> List[Tuple[str, Any, bool]]:
		"""
		Filter the list of members to always include init attributes unless the
		``:no-init-attribs:`` flag was given.

		:param members:
		:param want_all:
		"""  # noqa: D400

		attrib_names = (a.name for a in attr.fields(self.object) if a.init)

		no_init_attribs = self.options.get("no-init-attribs", False)

		def unskip_attrs(app, what, name, obj, skip, options):
			if skip and not no_init_attribs:
				return not (name in attrib_names)
			elif no_init_attribs and (name in attrib_names):
				return True
			return None

		listener_id = self.env.app.connect("autodoc-skip-member", unskip_attrs)
		members_ = super().filter_members(members, want_all)
		self.env.app.disconnect(listener_id)

		return members_

	def generate(
			self,
			more_content: Optional[Any] = None,
			real_modname: Optional[str] = None,
			check_module: bool = False,
			all_members: bool = False,
			) -> None:
		"""
		Generate reST for the object given by ``self.name``, and possibly for its members.

		:param more_content: Additional content to include in the reST output.
		:param real_modname: Module name to use to find attribute documentation.
		:param check_module: If :py:obj:`True`, only generate if the object is defined
			in the module name it is imported from.
		:param all_members: If :py:obj:`True`, document all members.
		"""

		if not self.parse_name():  # pragma: no cover
			# need a module to import
			unknown_module_warning(self)
			return None

		# now, import the module and get object to document
		if not self.import_object():
			return None  # pragma: no cover

		return super().generate(
				more_content=more_content,
				real_modname=real_modname,
				check_module=check_module,
				all_members=all_members,
				)


def setup(app: Sphinx) -> Dict[str, Any]:
	"""
	Setup :mod:`attr_utils.autoattrs`.

	:param app:

	.. versionadded:: 0.1.0
	"""

	# Hack to get the docutils tab size, as there doesn't appear to be any other way
	app.setup_extension("sphinx_toolbox.tweaks.tabsize")

	app.setup_extension("sphinx_toolbox.more_autosummary")

	app.add_autodocumenter(AttrsDocumenter)

	return {
			"version": __version__,
			"parallel_read_safe": True,
			}



#!/usr/bin/env python
#
#  pprint.py
r"""
Pretty printing functions.

This module monkeypatches `PrettyPrinter <https://prettyprinter.readthedocs.io>`_
to disable a potentially undesirable behaviour of its singledispatch feature,
where deferred types took precedence over resoled types.

It also changes the pretty print output for an :class:`enum.Enum` to be the same as the
:class:`~enum.Enum`\'s ``__repr__``.

The following functions are available:

"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Based on https://prettyprinter.readthedocs.io
#  Copyright © 2017, Tommi Kaikkonen.
#

# stdlib
import enum
import sys
from typing import Type, TypeVar

# 3rd party
import attr
import prettyprinter  # type: ignore
from prettyprinter.prettyprinter import _BASE_DISPATCH, pretty_dispatch, register_pretty  # type: ignore

__all__ = ["pretty_enum", "pretty_repr", "register_pretty"]

_T = TypeVar("_T")

prettyprinter.install_extras(["attrs"])


def is_registered(
		type,  # noqa: A002  # pylint: disable=redefined-builtin
		*,
		check_superclasses: bool = False,
		check_deferred: bool = True,
		register_deferred: bool = True,
		):
	if not check_deferred and register_deferred:
		raise ValueError("register_deferred may not be True when check_deferred is False")

	if type in pretty_dispatch.registry:
		return True

	if not check_superclasses:
		return False

	return pretty_dispatch.dispatch(type) is not _BASE_DISPATCH


# Resolve deferred names and prevent them being used again.
prettyprinter.is_registered(type(_T), check_superclasses=True, check_deferred=True, register_deferred=True)
prettyprinter.is_registered = is_registered
sys.modules["prettyprinter.prettyprinter"].is_registered = is_registered  # type: ignore


@register_pretty(enum.EnumMeta)
@register_pretty(enum.Enum)
@register_pretty(enum.IntEnum)
@register_pretty(enum.Flag)
@register_pretty(enum.IntFlag)
def pretty_enum(value, ctx) -> str:
	r"""
	Pretty-prints the given :class:`~enum.Enum`.
	"""

	return repr(value)


def pretty_repr(obj: Type):
	"""
	Add a pretty-printed ``__repr__`` function to the decorated attrs class.

	.. code-block:: python

		>>> import attr
		>>> from attr_utils.pprinter import pretty_repr

		>>> @pretty_repr
		... @attr.s
		... class Person(object):
		... 	name = attr.ib()

		>>> repr(Person(name="Bob"))
		Person(name='Bob')

	:param obj:
	"""

	if attr.has(obj):

		def __repr__(self) -> str:
			return prettyprinter.pformat(self)

		__repr__.__doc__ = f"Return a string representation of the :class:`~.{obj.__name__}`."

		obj.__repr__ = __repr__  # type: ignore
		obj.__repr__.__qualname__ = f"{obj.__name__}.__repr__"
		obj.__repr__.__module__ = obj.__module__

	return obj



#!/usr/bin/env python
#
#  annotations.py
"""
Add type annotations to the ``__init__`` of an attrs_ class.

Since :pull:`363 <python-attrs/attrs>` attrs has
populated the ``__init__.__annotations__`` based on the types of attributes.
However, annotations were deliberately omitted when converter functions were used.
This module attempts to generate the annotations for use in Sphinx documentation,
even when converter functions *are* used, based on the following assumptions:

* If the converter function is a Python ``type``, such as :class:`str`, :class:`int`,
  or :class:`list`, the type annotation will be that type.
  If the converter and the type annotation refer to the same type
  (e.g. :class:`list` and :class:`typing.List`) the type annotation will be used.

* If the converter function has an annotation for its first argument, that annotation is used.

* If the converter function is not annotated, the type of the attribute will be used.

The annotation can also be provided via the ``'annotation'`` key in the
`metadata dict <https://www.attrs.org/en/stable/examples.html#metadata>`_.
If you prefer you can instead provide this as a keyword argument to :func:`~.attrib`
which will construct the metadata dict and call :func:`attr.ib` for you.

.. _attrs: https://www.attrs.org/en/stable/

.. versionchanged:: 0.2.0

	Improved support for container types.


.. attention::

	Due to changes in the :mod:`typing` module :mod:`~attr_utils.annotations`
	is only officially supported on Python 3.7 and above.

Examples
---------------

**Library Usage:**

.. code-block:: python

	def my_converter(arg: List[Dict[str, Any]]):
		return arg


	def untyped_converter(arg):
		return arg


	@attr.s
	class SomeClass:
		a_string: str = attr.ib(converter=str)
		custom_converter: Any = attr.ib(converter=my_converter)
		untyped: Tuple[str, int, float] = attr.ib(converter=untyped_converter)
		annotated: List[str] = attr.ib(converter=list, metadata={"annotation": Sequence[str]})

	add_attrs_annotations(SomeClass)

	print(SomeClass.__init__.__annotations__)
	# {
	#	'return': None,
	#	'a_string': <class 'str'>,
	#	'custom_converter': typing.List[typing.Dict[str, typing.Any]],
	#	'untyped': typing.Tuple[str, int, float],
	#	}

**Sphinx documentation**:

	.. literalinclude:: ../../attr_utils/annotations.py
		:tab-width: 4
		:pyobject: AttrsClass

The ``parse_occupations`` looks like:

	.. literalinclude:: ../../attr_utils/annotations.py
		:tab-width: 4
		:pyobject: parse_occupations

The Sphinx output looks like:

	.. autoclass:: attr_utils.annotations.AttrsClass
		:members:
		:no-special-members:


API Reference
---------------

"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import inspect
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Type, Union

# 3rd party
import attr
from typing_inspect import get_origin  # type: ignore

# this package
from attr_utils import __version__

if TYPE_CHECKING:
	# 3rd party
	from sphinx.application import Sphinx

__all__ = ["add_init_annotations", "attrib", "attr_docstring_hook", "setup"]


def add_init_annotations(obj: Callable) -> Callable:
	"""
	Add type annotations to the ``__init__`` of an attrs_ class.

	.. _attrs: https://www.attrs.org/en/stable/
	"""

	if not attr.has(obj):  # type: ignore
		return obj

	annotations: Dict[str, Optional[Type]] = {"return": None}

	attrs = attr.fields(obj)  # type: ignore

	for a in attrs:
		arg_name = a.name.lstrip('_')

		if a.init is True and a.type is not None:
			if a.converter is None:
				annotations[arg_name] = a.type
			else:

				if "annotation" in a.metadata:
					annotations[arg_name] = a.metadata["annotation"]

				elif isinstance(a.converter, type):
					if a.converter is get_origin(a.type):
						annotations[arg_name] = a.type
					else:
						annotations[arg_name] = a.converter

				else:
					signature = inspect.signature(a.converter)
					arg_type = next(iter(signature.parameters.items()))[1].annotation
					if arg_type is inspect.Signature.empty:
						annotations[arg_name] = a.type
					else:
						annotations[arg_name] = arg_type

	if hasattr(obj.__init__, "__annotations__"):  # type: ignore
		obj.__init__.__annotations__.update(annotations)  # type: ignore
	else:  # pragma: no cover
		obj.__init__.__annotations__ = annotations  # type: ignore

	return obj


def attrib(
		default=attr.NOTHING,
		validator=None,
		repr: bool = True,  # noqa: A002  # pylint: disable=redefined-builtin
		hash=None,  # noqa: A002  # pylint: disable=redefined-builtin
		init=True,
		metadata=None,
		annotation: Union[Type, object] = attr.NOTHING,
		converter=None,
		factory=None,
		kw_only: bool = False,
		eq=None,
		order=None,
		**kwargs,
		):
	"""
	Wrapper around :func:`attr.ib` which supports the ``annotation``
	keyword argument for use by :func:`~.add_init_annotations`.

	:param metadata: The type to add to ``__init__.__annotations__``, if different from
		that the type taken as input to the converter function or the type hint of the attribute.

	See the documentation for :func:`attr.ib` for descriptions of the other arguments.

	.. versionadded:: 0.2.0
	"""  # noqa: D400

	if annotation is not attr.NOTHING:
		if metadata is None:
			metadata = {}

		metadata["annotation"] = annotation

	return attr.ib(
			default=default,
			validator=validator,
			repr=repr,
			hash=hash,
			init=init,
			metadata=metadata,
			converter=converter,
			factory=factory,
			kw_only=kw_only,
			eq=eq,
			order=order,
			**kwargs,
			)


def attr_docstring_hook(obj: Any) -> Any:
	"""
	Hook for :mod:`sphinx_toolbox.more_autodoc.typehints` to add annotations to the ``__init__`` of
	attrs_ classes.

	.. _attrs: https://www.attrs.org/en/stable/

	:param obj: The object being documented.
	"""  # noqa D400

	if callable(obj):

		if inspect.isclass(obj):
			obj = add_init_annotations(obj)

	return obj


def setup(app: "Sphinx") -> Dict[str, Any]:
	"""
	Sphinx extension to populate ``__init__.__annotations__`` for attrs_ classes.

	.. _attrs: https://www.attrs.org/en/stable/

	:param app:

	:return:
	"""

	# 3rd party
	from sphinx_toolbox.more_autodoc.typehints import docstring_hooks  # nodep

	docstring_hooks.append((attr_docstring_hook, 50))

	app.setup_extension("sphinx_toolbox.more_autodoc.typehints")

	return {
			"version": __version__,
			"parallel_read_safe": True,
			}


################################
# Demo
################################


def parse_occupations(occupations: Iterable[str]) -> Iterable[str]:  # pragma: no cover
	if isinstance(occupations, str):
		return [x.strip() for x in occupations.split(',')]
	else:
		return [str(x) for x in occupations]


@attr.s
class AttrsClass:
	"""
	Example of using :func:`~.add_init_annotations` for attrs_ classes with Sphinx documentation.

	.. _attrs: https://www.attrs.org/en/stable/

	:param name: The name of the person.
	:param age: The age of the person.
	:param occupations: The occupation(s) of the person.
	"""

	name: str = attr.ib(converter=str)
	age: int = attr.ib(converter=int)
	occupations: List[str] = attr.ib(converter=parse_occupations)



#!/usr/bin/env python3
#
#  __init__.py
"""
Utilities to augment attrs.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020 Dominic Davis-Foster"

__license__: str = "MIT License"
__version__: str = "0.5.5"
__email__: str = "dominic@davis-foster.co.uk"



#!/usr/bin/env python
#
#  mypy_plugin.py
"""
Plugin for `mypy <https://github.com/python/mypy>`_ which adds support for attr_utils.

.. versionadded:: 0.4.0

To use this plugin, add the following to your
`mypy configuration file <https://mypy.readthedocs.io/en/stable/config_file.html>`_:

.. code-block:: ini

	[mypy]
	plugins=attr_utils.mypy_plugin

See https://mypy.readthedocs.io/en/stable/extending_mypy.html#configuring-mypy-to-use-plugins
for more information.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Based on https://github.com/python/mypy/issues/5719
#  and https://gitter.im/python/typing?at=5e078653eac8d1511e737d8c
#
#  Also based on mypy
#  https://github.com/python/mypy
#  Copyright (c) 2015-2019 Jukka Lehtosalo and contributor
#  Licensed under the terms of the MIT license.
#

# stdlib
from typing import Callable, List, MutableMapping, Optional

# 3rd party
from mypy.nodes import (  # nodep
		ARG_OPT,
		ARG_POS,
		MDEF,
		Argument,
		Block,
		ClassDef,
		FuncDef,
		PassStmt,
		SymbolTableNode,
		Var
		)
from mypy.plugin import ClassDefContext, Plugin, SemanticAnalyzerPluginInterface  # nodep
from mypy.plugins.common import add_method_to_class  # nodep
from mypy.semanal_shared import set_callable_name  # nodep
from mypy.types import AnyType, CallableType, Instance, Type, TypeOfAny, TypeType, TypeVarDef  # nodep
from mypy.typevars import fill_typevars  # nodep
from mypy.util import get_unique_redefinition_name  # nodep

__all__ = ["attr_utils_serialise_serde", "AttrUtilsPlugin", "add_classmethod_to_class", "plugin"]

#: Registry mapping decorator full names to the callable that handles the methods added by the decorator.
decorator_registry: MutableMapping[str, Callable[[ClassDefContext], None]] = {}


def attr_utils_serialise_serde(cls_def_ctx: ClassDefContext):
	"""
	Handles :func:attr_utils.serialise.serde`.

	:param cls_def_ctx:
	"""

	info = cls_def_ctx.cls.info

	# https://gitter.im/python/typing?at=5e078653eac8d1511e737d8c
	str_type = cls_def_ctx.api.named_type("__builtins__.str")
	bool_type = cls_def_ctx.api.named_type("__builtins__.bool")
	implicit_any = AnyType(TypeOfAny.special_form)
	mapping = cls_def_ctx.api.lookup_fully_qualified_or_none("typing.Mapping")
	mutable_mapping = cls_def_ctx.api.lookup_fully_qualified_or_none("typing.MutableMapping")
	mapping_str_any_type = Instance(mapping.node, [str_type, implicit_any])  # type: ignore
	mutable_mapping_str_any_type = Instance(mutable_mapping.node, [str_type, implicit_any])  # type: ignore
	# # maybe_mapping_str_any = UnionType.make_union([typ, NoneType()])(mapping_str_any_type)
	decorated_class_instance = Instance(
			cls_def_ctx.api.lookup_fully_qualified_or_none(cls_def_ctx.cls.fullname).node,  # type: ignore
			[],
			)

	if "to_dict" not in info.names:
		add_method_to_class(
				api=cls_def_ctx.api,
				cls=cls_def_ctx.cls,
				name="to_dict",
				args=[Argument(Var("convert_values", bool_type), bool_type, None, ARG_OPT)],
				return_type=mutable_mapping_str_any_type,
				)

	if "from_dict" not in info.names:
		add_classmethod_to_class(
				api=cls_def_ctx.api,
				cls=cls_def_ctx.cls,
				name="from_dict",
				args=[Argument(Var('d', mapping_str_any_type), mapping_str_any_type, None, ARG_POS)],
				return_type=decorated_class_instance,
				cls_type=TypeType(decorated_class_instance),
				)


#
# def attr__make_attrs(cls_def_ctx: ClassDefContext):
# 	"""
# 	Handles :func:attr.ib`.
#
# 	:param cls_def_ctx:
# 	"""
#
# 	info = cls_def_ctx.cls.info
#
# 	list_ = cls_def_ctx.api.lookup_fully_qualified_or_none('typing.List')
# 	attribute = cls_def_ctx.api.lookup_fully_qualified_or_none('attr.Attribute')
# 	#
# 	# if "__attrs_attrs__" not in info.names:
# 	# 	info.names["__attrs_attrs__"] = SymbolTableNode(
# 	# 			MDEF,
# 	# 			Var("__attrs_attrs__", Instance(list_.node, [attribute])),  # type: ignore
# 	# 			plugin_generated=True,
# 	# 			)

decorator_registry["attr_utils.serialise.serde"] = attr_utils_serialise_serde
# decorator_registry["attr._make.attrs"] = attr__make_attrs
# decorator_registry["attr.s"] = attr__make_attrs
# decorator_registry["attr.attrs"] = attr__make_attrs


class AttrUtilsPlugin(Plugin):
	"""
	Plugin for `mypy <https://github.com/python/mypy>`_ which adds support for attr_utils.
	"""

	def get_class_decorator_hook(self, fullname: str) -> Optional[Callable[[ClassDefContext], None]]:
		"""
		Allows mypy to handle decorators that add extra methods to classes.

		:param fullname: The full name of the decorator.
		"""

		return decorator_registry.get(fullname, None)


def add_classmethod_to_class(
		api: SemanticAnalyzerPluginInterface,
		cls: ClassDef,
		name: str,
		args: List[Argument],
		return_type: Type,
		cls_type: Optional[Type] = None,
		tvar_def: Optional[TypeVarDef] = None,
		) -> None:
	"""
	Adds a new classmethod to a class definition.
	"""
	info = cls.info

	# First remove any previously generated methods with the same name
	# to avoid clashes and problems in the semantic analyzer.
	if name in info.names:
		sym = info.names[name]
		if sym.plugin_generated and isinstance(sym.node, FuncDef):
			cls.defs.body.remove(sym.node)

	cls_type = cls_type or fill_typevars(info)
	function_type = api.named_type("__builtins__.function")

	args = [Argument(Var("cls"), cls_type, None, ARG_POS)] + args
	arg_types, arg_names, arg_kinds = [], [], []
	for arg in args:
		assert arg.type_annotation, "All arguments must be fully typed."
		arg_types.append(arg.type_annotation)
		arg_names.append(arg.variable.name)
		arg_kinds.append(arg.kind)

	signature = CallableType(arg_types, arg_kinds, arg_names, return_type, function_type)
	if tvar_def:
		signature.variables = [tvar_def]

	func = FuncDef(name, args, Block([PassStmt()]))
	func.is_class = True
	func.info = info
	func.type = set_callable_name(signature, func)
	func._fullname = info.fullname + '.' + name
	func.line = info.line

	# NOTE: we would like the plugin generated node to dominate, but we still
	# need to keep any existing definitions so they get semantically analyzed.
	if name in info.names:
		# Get a nice unique name instead.
		r_name = get_unique_redefinition_name(name, info.names)
		info.names[r_name] = info.names[name]

	info.names[name] = SymbolTableNode(MDEF, func, plugin_generated=True)
	info.defn.defs.body.append(func)


def plugin(version: str):
	"""
	Entry point to :mod:`attr_utils.mypy_plugin`.

	:param version: The current mypy version.
	"""

	# ignore version argument if the plugin works with all mypy versions.
	return AttrUtilsPlugin
